import re
from typing import TypedDict, Iterator


class CanvasResource(TypedDict):
    type: str
    id: str
    data: dict
    content_path: str


class ResourceInfo(TypedDict):
    id: str


class AnnouncementInfo(ResourceInfo):
    id: str
    url: str | None
    uri: str | None  # for course-link
    title: str  # for course-link


class CourseSettingsInfo(ResourceInfo):
    id: str


class AssignmentInfo(ResourceInfo):
    id: str
    url: str | None
    uri: str | None  # for course-link
    title: str  # for course-link text


class FileInfo(ResourceInfo):
    id: str
    uri: str
    title: str  # for course-link


class AssignmentGroupInfo(ResourceInfo):
    id: str


class ModuleInfo(ResourceInfo):
    id: str
    title: str  # for course-link
    uri: str
    url: str


class ModuleItemInfo(ResourceInfo):
    id: str
    uri: str
    url: str


class OverrideInfo(ResourceInfo):
    id: str


class PageInfo(ResourceInfo):
    id: str
    page_url: str  # for module item
    uri: str  # for course-link
    url: str | None
    title: str  # for course-link text


class QuizInfo(ResourceInfo):
    id: str
    uri: str  # for course-link
    url: str | None
    title: str  # for course-link text


class SyllabusInfo(ResourceInfo):
    id: str
    uri: str
    url: str
    title: str  # for course-link title


class CourseSettings(TypedDict):
    name: str
    code: str
    image: str


class FileData(TypedDict):
    path: str
    canvas_folder: str | None
    lock_at: str | None
    unlock_at: str | None


class ZipFileData(TypedDict):
    zip_file_name: str
    content_folder: str
    additional_files: list[str] | None
    exclude_pattern: str | None
    priority_folder: str | None
    canvas_folder: str | None


class SyllabusData(TypedDict):
    content: str


def iter_keys(text: str) -> Iterator[tuple[str, str, str, str]]:
    for match in re.finditer(fr'__@@([^|]+)\|\|([^|]+)\|\|([^@]+)@@__', text):
        yield match.group(0), *match.groups()


def get_key(rtype: str, rid: str, field: str):
    return f'__@@{rtype}||{rid}||{field}@@__'


class ResourceManager(dict[tuple[str, str], CanvasResource]):

    def add_resource(self, resource: CanvasResource, field: str = None) -> str:
        rtype = resource['type']
        rid = resource['id']
        self[rtype, rid] = resource
        return get_key(rtype, rid, field) if field else None



from canvasapi.course import Course

from ..resources import AnnouncementInfo


def deploy_announcement(course: Course, announcement_info: dict) -> tuple[AnnouncementInfo, None]:
    announcement_id = announcement_info["canvas_id"]

    if announcement_id:
        canvas_announcements = course.get_discussion_topics(course_id=course.id, only_announcements=True)
        canvas_announcement = next(
            (a for a in canvas_announcements if a.id == announcement_id), None
        )
        canvas_announcement.update(**announcement_info)

    else:
        canvas_announcement = course.create_discussion_topic(**announcement_info)

    announcement_object_info: AnnouncementInfo = {
        'id': canvas_announcement.id,
        'title': canvas_announcement.title,
        'uri': f'/courses/{course.id}/discussion_topics/{canvas_announcement.id}',

        # Following fields have been observed to be missing in some cases
        'url': canvas_announcement.html_url if hasattr(canvas_announcement, 'html_url') else None
    }

    return announcement_object_info, None



from typing import Any

from bs4 import Tag

from .attributes import Attribute, parse_bool, parse_dict, parse_list, parse_settings, parse_int
from ..resources import ResourceManager, get_key, CanvasResource
from ..error_helpers import format_tag, get_file_path
from ..processing_context import get_current_file


def _parse_module_list(text: str) -> list[str]:
    modules = parse_list(text)
    return [get_key('module', module, 'id') for module in modules]


class ModuleTagProcessor:
    def __init__(self, resource_manager: ResourceManager):
        self._resources = resource_manager
        self._previous_module = None  # The id of the previous module
        self._previous_module_item = None  # The id of the previous module item
        self._previous_module_position = 1

    _module_item_type_casing = {
        "file": "File",
        "page": "Page",
        "discussion": "Discussion",
        "assignment": "Assignment",
        "quiz": "Quiz",
        "subheader": "SubHeader",
        "externalurl": "ExternalUrl",
        "externaltool": "ExternalTool"
    }

    def __call__(self, module_tag: Tag):
        fields = [
            Attribute('id', ignore=True),
            Attribute('title', required=True, new_name='name'),
            Attribute('position'),
            Attribute('published', parser=parse_bool),
            Attribute('previous-module'),
            Attribute('prerequisite_module_ids', parser=_parse_module_list)
        ]

        module_data = parse_settings(module_tag, fields)

        module_data['_comments'] = {
            'previous_module': ''
        }

        if self._previous_module is not None:
            # adding a reference to the previous module ensures this module
            #  is created after the previous one, thus preserving their
            #  relative ordering
            module_data['_comments']['previous_module'] = get_key('module', self._previous_module, 'id')

        if prev_mod := module_data.get('previous-module'):
            module_data['_comments']['previous_module'] = get_key('module', prev_mod, 'id')

        module_id = module_tag.get('id', module_data['name'])
        self._previous_module = module_id

        self._resources.add_resource(CanvasResource(
            type='module',
            id=module_id,
            data=module_data,
            content_path=str(get_current_file().resolve())
        ))

        self._previous_module_item = None
        self._previous_module_position = 1
        for item_tag in module_tag.find_all('item'):
            self._parse_module_item(module_id, item_tag)

    def _parse_module_item(self, module_rid: str, tag: Tag):
        fields = [
            Attribute('type', ignore=True),
            Attribute('position', parser=parse_int),
            Attribute('indent', parser=parse_int),
            Attribute('new_tab', True, parse_bool),
            Attribute('completion_requirement', parser=parse_dict),
            Attribute('iframe'),
            Attribute('published', parser=parse_bool),
        ]

        rtype = self._module_item_type_casing[tag['type'].lower()]
        item: dict[str, Any] = {
            'type': rtype
        }

        if rtype == 'ExternalUrl':
            fields.extend([
                Attribute('external_url', required=True),
                Attribute('title'),
                Attribute('id')
            ])
            item.update(parse_settings(tag, fields))
            if 'title' not in item:
                item['title'] = item['external_url']
            if 'id' not in item:
                item['id'] = item['title']

        elif rtype == 'SubHeader':
            fields.extend([
                Attribute('title', required=True),
                Attribute('id'),
            ])
            item.update(parse_settings(tag, fields))
            if 'id' not in item:
                item['id'] = item['title']

        elif rtype in ['Page', 'Quiz', 'Assignment', 'File']:
            fields.extend([
                Attribute('content_id', ignore=True),
                Attribute('title')
            ])

            if not (rid := tag.get('content_id')):
                raise ValueError(
                    f'Module "{rtype}" item must have "content_id" @ {format_tag(tag)}\n  in {get_file_path(tag)}')

            item.update(parse_settings(tag, fields))
            if rtype == 'Page':
                item['page_url'] = get_key('page', rid, 'page_url')
            else:
                item['content_id'] = get_key(rtype.lower(), rid, 'id')
            item['id'] = rid

        else:
            raise NotImplementedError(
                f'Unrecognized module item type "{rtype}" @ {format_tag(tag)}\n  in {get_file_path(tag)}')

        # Namespace each module item ID to the module
        # Otherwise, a resource can only be linked to a single module
        item['id'] = f'{module_rid}|{item["id"]}'
        item['module_id'] = get_key('module', module_rid, 'id')

        if prev_pos := item.get('position'):
            self._previous_module_position = prev_pos
        else:
            item['position'] = self._previous_module_position
            self._previous_module_position += 1

        item['_comments'] = {
            'previous_module_item':
                get_key('module_item', self._previous_module_item, 'id')
                if self._previous_module_item is not None
                else ''
        }
        self._previous_module_item = item['id']

        self._resources.add_resource(CanvasResource(
            type='module_item',
            id=item['id'],
            data=item,
            content_path=str(get_current_file().resolve())
        ))



from canvasapi.course import Course

from ..resources import SyllabusData, SyllabusInfo


def deploy_syllabus(course: Course, data: SyllabusData) -> tuple[SyllabusInfo, None]:
    course.update(course={'syllabus_body': data['content']})

    syllabus_object_info: SyllabusInfo = {
        'id': str(course.id),
        'title': 'Syllabus',
        'uri': f'/courses/{course.id}/assignments/syllabus',
        'url': f'{course.canvas._Canvas__requester.original_url}/courses/{course.id}/assignments/syllabus'
    }

    return syllabus_object_info, None



from collections import deque

from ..our_logging import get_logger

logger = get_logger()


def tarjan_scc(graph: dict[tuple[str, str], list[tuple[str, str]]]) -> list[list[tuple[str, str]]]:
    """Find strongly connected components (cycles) using Tarjan's algorithm."""
    index_counter = [0]
    stack = []
    lowlink = {}
    index = {}
    on_stack = set()
    sccs = []

    def strongconnect(node):
        index[node] = index_counter[0]
        lowlink[node] = index_counter[0]
        index_counter[0] += 1
        stack.append(node)
        on_stack.add(node)

        for successor in graph.get(node, []):
            if successor not in index:
                strongconnect(successor)
                lowlink[node] = min(lowlink[node], lowlink[successor])
            elif successor in on_stack:
                lowlink[node] = min(lowlink[node], index[successor])

        if lowlink[node] == index[node]:
            scc = []
            while True:
                w = stack.pop()
                on_stack.remove(w)
                scc.append(w)
                if w == node:
                    break
            sccs.append(scc)

    for node in graph.keys():
        if node not in index:
            strongconnect(node)

    return sccs


def kahns_topological_sort(graph: dict[tuple[str, str], list[tuple[str, str]]]) -> list[tuple[str, str]]:
    """Topological sort using Kahn's algorithm (dependencies first)."""
    in_degree = {node: 0 for node in graph}
    for deps in graph.values():
        for dep in deps:
            if dep in graph:
                in_degree[dep] += 1

    queue = deque([node for node, degree in in_degree.items() if degree == 0])
    result = []

    while queue:
        node = queue.popleft()
        result.append(node)

        for dep in graph.get(node, []):
            if dep in in_degree:
                in_degree[dep] -= 1
                if in_degree[dep] == 0:
                    queue.append(dep)

    if len(result) != len(graph):
        raise ValueError("Internal error: acyclic graph still has cycles")

    return result[::-1]


def linearize_dependencies(
        graph: dict[tuple[str, str], list[tuple[str, str]]],
        shell_deployers: list[str]
) -> list[tuple[tuple[str, str], bool]]:
    """
    Linearize dependencies with cycle breaking via shell deployments.
    Example: A→B→C→A returns [((page, C),True), ((page, B), False), ((page, A), False), ((page, C), False)]

    Note: Only certain resource types support shell deployments (assignment, page, quiz).
    Other types (syllabus, file, etc.) have deterministic IDs and don't need shell deployments.
    """

    sccs = tarjan_scc(graph)

    cycle_breakers = set()
    scc_map = {}

    for scc in sccs:
        if len(scc) > 1:
            # All nodes in an SCC that are depended upon by other nodes in the SCC
            # need to be cycle breakers (deployed as shells first)
            for node in scc:
                scc_map[node] = scc
                # Check if any other node in the SCC depends on this node
                # and that this node supports shell deployment
                if node[0] in shell_deployers:
                    for other_node in scc:
                        if other_node != node and node in graph.get(other_node, []):
                            cycle_breakers.add(node)
                            break

    for node, deps in graph.items():
        # Self-loops only need cycle breaking if the type supports shell deployment
        if node in deps and node[0] in shell_deployers:
            cycle_breakers.add(node)

    acyclic_graph = {}
    for node, deps in graph.items():
        if node in scc_map:
            # Remove all edges within the same SCC to break cycles
            acyclic_graph[node] = [d for d in deps if d not in scc_map[node]]
        elif node in cycle_breakers:
            # Self-loop case
            acyclic_graph[node] = [d for d in deps if d != node]
        else:
            acyclic_graph[node] = deps

    topo_order = kahns_topological_sort(acyclic_graph)

    # Shell deployments for cycle breakers must happen FIRST (before everything else)
    # so that other resources can reference them. Then the full deployment happens
    # in the natural topological order.
    result = [(node, True) for node in topo_order if node in cycle_breakers]
    result.extend((node, False) for node in topo_order)

    return result


if __name__ == '__main__':
    dependency_dict = {
        'A': ['B'],
        'B': ['C', 'D'],
        'C': [],
        'D': ['C']
    }

    # expect C, D, B, A
    order = linearize_dependencies(dependency_dict)
    print(order)



from canvasapi.course import Course

from ..resources import AssignmentGroupInfo


def deploy_group(course: Course, group_data: dict) -> tuple[AssignmentGroupInfo, None]:
    group_id = group_data["canvas_id"]

    if group_id:
        group = course.get_assignment_group(group_id)
    else:
        group = course.create_assignment_group(name=group_data["name"])

    group.edit(**group_data)
    course.update(course={
        'apply_assignment_group_weights': True,
    })

    group_object_info: AssignmentGroupInfo = {
        'id': group.id
    }

    return group_object_info, None



from bs4 import Tag

from mdxcanvas.resources import ResourceManager, CanvasResource, SyllabusData
from mdxcanvas.util import retrieve_contents
from mdxcanvas.processing_context import get_current_file


class SyllabusTagProcessor:
    def __init__(self, resources: ResourceManager):
        self._resources = resources

    def __call__(self, tag: Tag):
        syllabus = CanvasResource(
            type='syllabus',
            id='syllabus',
            data=SyllabusData(content=retrieve_contents(tag)),
            content_path=str(get_current_file().resolve())
        )
        self._resources.add_resource(syllabus)



from pathlib import Path
from typing import Callable

from bs4 import Tag

from ..error_helpers import format_tag, get_file_path, validate_required_attribute
from ..our_logging import get_logger
from ..processing_context import FileContext, get_current_file
from ..resources import ResourceManager, FileData, ZipFileData, CanvasResource, get_key
from ..util import parse_soup_from_xml
from ..xml_processing.attributes import parse_bool, get_tag_path

logger = get_logger()


def make_course_settings_preprocessor(parent: Path, resources: ResourceManager):
    def process_course_settings(tag: Tag):
        name = tag.get('name')
        course_code = tag.get('code')
        image_path = tag.get('image')

        if not (any([name, course_code, image_path])):
            raise ValueError(
                f"Course settings tag must have at least one of: name, code, or image @ {format_tag(tag)}\n  in {get_file_path(tag)}")

        image_resource_key = None
        if image_path:
            image_path = (parent / image_path).resolve().absolute()
            if not image_path.is_file():
                raise ValueError(
                    f"Course image file not found @ {format_tag(tag)}\n  Image path: {image_path}\n  in {get_file_path(tag)}")

            file = CanvasResource(
                type='file',
                id=image_path.name,
                data=FileData(
                    path=str(image_path),
                    canvas_folder=tag.get('canvas_folder', None)
                ),
                content_path=str(get_current_file().resolve())
            )
            image_resource_key = resources.add_resource(file, 'id')

        course_settings = CanvasResource(
            type='course_settings',
            id='',
            data={
                'name': name,
                'code': course_code,
                'image': image_resource_key
            },
            content_path=str(get_current_file().resolve())
        )
        resources.add_resource(course_settings, 'name')

    return process_course_settings


def make_image_preprocessor(parent: Path, resources: ResourceManager):
    def process_image(tag: Tag):
        # TODO - handle b64-encoded images

        src = validate_required_attribute(tag, 'src', 'img')
        if src.startswith('http') or src.startswith('__@@'):
            # No changes necessary
            return

        # Assume it's a local file
        src_path = (parent / src).resolve().absolute()
        if not src_path.is_file():
            raise ValueError(
                f"Image file not found @ {format_tag(tag)}\n  File path: {src_path}\n  in {get_file_path(tag)}")
        src = src_path

        file = CanvasResource(
            type='file',
            id=src.name,
            data=FileData(
                path=str(src),
                canvas_folder=tag.get('canvas_folder', None)
            ),
            content_path=str(get_current_file().resolve())
        )
        tag['src'] = resources.add_resource(file, 'uri') + '/preview'

    return process_image


def make_file_anchor_tag(resource_key: str, filename: str, **kwargs):
    attrs = {
        **kwargs,
        'href': f'{resource_key}?wrap=1',
        'class': 'instructure_file_link inline_disabled',
        'title': filename,
        'target': '_blank',
        'rel': 'noopener noreferrer'
    }

    new_tag = Tag(name='a', attrs=attrs)
    new_tag.string = filename

    return new_tag


def make_file_preprocessor(parent: Path, resources: ResourceManager):
    def process_file(tag: Tag):
        path_value = validate_required_attribute(tag, 'path', 'file')
        attrs = tag.attrs
        path = (parent / path_value).resolve().absolute()
        if not path.is_file():
            raise ValueError(f"File not found @ {format_tag(tag)}\n  File path: {path}\n  in {get_file_path(tag)}")
        # Only pop the path after validation succeeds
        attrs.pop('path')
        file = CanvasResource(
            type='file',
            id=path.name,
            data=FileData(
                path=str(path),
                canvas_folder=attrs.get('canvas_folder', None),
                unlock_at=attrs.get('unlock_at', None),
                lock_at=attrs.get('lock_at', None)
            ),
            content_path=str(get_current_file().resolve())
        )
        resource_key = resources.add_resource(file, 'uri')
        new_tag = make_file_anchor_tag(resource_key, path.name, **tag.attrs)

        tag.replace_with(new_tag)

    return process_file


def make_zip_preprocessor(parent: Path, resources: ResourceManager):
    def process_zip(tag: Tag):
        content_folder = validate_required_attribute(tag, 'path', 'zip')

        name = tag.get("name")
        if not name:
            name = (
                       content_folder
                       .replace('.', '')
                       .replace('/', '-')
                       .strip('-')
                   ) + '.zip'

        content_folder_path = (parent / content_folder).resolve().absolute()
        if not content_folder_path.exists():
            raise ValueError(
                f"Folder not found @ {format_tag(tag)}\n  Folder path: {content_folder_path}\n  in {get_file_path(tag)}")
        content_folder = str(content_folder_path)

        additional_files = tag.get("additional_files")
        if additional_files:
            additional_files = [str((parent / file).resolve().absolute()) for file in additional_files.split(',')]

        priority_folder = tag.get("priority_path")
        if priority_folder:
            priority_folder = str((parent / priority_folder).resolve().absolute())

        exclude_pattern = tag.get("exclude")

        file = CanvasResource(
            type='zip',
            id=name,
            data=ZipFileData(
                zip_file_name=name,
                content_folder=content_folder,
                additional_files=additional_files,
                exclude_pattern=exclude_pattern,
                priority_folder=priority_folder,
                canvas_folder=tag.get('canvas_folder')
            ),
            content_path=str(get_current_file().resolve())
        )

        resource_key = resources.add_resource(file, 'uri')

        new_tag = make_file_anchor_tag(resource_key, name)
        tag.replace_with(new_tag)

    return process_zip


def _parse_slice(field: str) -> slice:
    """
    Parse a 1-based, inclusive slice
    So, the slice should match the line numbers shown in your IDE
    """
    tokens = field.split(':')
    tokens = [
        int(token) if token else None
        for token in tokens
    ]

    tokens[0] -= 1  # make it 1-based

    if len(tokens) == 1:  # e.g. "3"
        tokens.append(None)

    # Tokens[1] +1 for inclusive, -1 for one-based, net: 0

    return slice(tokens[0], tokens[1])


def make_include_preprocessor(
        parent_folder: Path,
        process_file: Callable
):
    def process_include(tag: Tag):
        imported_filename = tag.get('path')
        imported_file = (parent_folder / imported_filename).resolve()
        args_file_path = tag.get('args', None)

        # Check if the included file exists first
        if not imported_file.exists():
            containing_file = get_current_file() or "unknown file"
            raise FileNotFoundError(
                f"Include file not found @ {format_tag(tag)}\n"
                f"  in {containing_file}"
            )

        args_file = None
        if args_file_path is not None:
            args_file = (parent_folder / args_file_path).resolve().absolute()

            # Check if args file exists
            if not args_file.exists():
                containing_file = get_current_file() or "unknown file"
                raise FileNotFoundError(
                    f"Args file not found @ {format_tag(tag)}\n"
                    f"  in {containing_file}"
                )

        # Track the included file in context for error messages
        with FileContext(imported_file):
            imported_raw_content = imported_file.read_text(encoding='utf-8')
            suffixes = imported_file.suffixes

            lines = tag.get('lines', '')
            if lines:
                grab = _parse_slice(lines)
                imported_raw_content = '\n'.join(imported_raw_content.splitlines()[grab])

            if parse_bool(tag.get('fenced', 'false')):
                imported_raw_content = f'```{imported_file.suffix.lstrip(".")}\n{imported_raw_content}\n```\n'
                suffixes = suffixes + ['.md']

            imported_html = process_file(
                imported_file.parent,
                imported_raw_content,
                suffixes,
                args_file=args_file
            )

            use_div = parse_bool(tag.get('usediv', 'true'))

            include_result = parse_soup_from_xml(imported_html)

            if not use_div:
                tag.replace_with(include_result)

            else:
                new_tag = Tag(name='div')
                new_tag['data-source'] = str(imported_file)
                if lines:
                    new_tag['data-lines'] = lines
                new_tag.extend(include_result)
                tag.replace_with(new_tag)

    return process_include


def make_link_preprocessor():
    def process_link(tag: Tag):
        link_type = validate_required_attribute(tag, 'type', 'course-link')
        # TODO: Canvas supports `course navigation` links, do we?
        if link_type not in ['syllabus', 'page', 'assignment', 'quiz', 'announcement', 'discussion', 'module', 'file']:
            raise ValueError(f'Invalid course-link type "{link_type}" @ {format_tag(tag)}\n  in {get_file_path(tag)}')

        link_rid = validate_required_attribute(tag, 'id', 'course-link')

        new_tag = Tag(name='a')
        new_tag['href'] = get_key(link_type, link_rid, 'uri')
        link_text = tag.string.strip() if tag.string is not None else ''
        if not link_text:
            link_text = get_key(link_type, link_rid, 'title')
        new_tag.string = link_text
        tag.replace_with(new_tag)

    return process_link


def make_markdown_page_preprocessor(
        parent_folder: Path,
        process_file: Callable
):
    def process_markdown_page(tag: Tag):
        content_path = validate_required_attribute(tag, 'path', 'md-page')
        content_path_obj = Path(parent_folder) / content_path

        page_title = tag.get('title')
        if page_title is None:
            # Only read file if it exists for getting the title
            if content_path_obj.exists():
                first_line = content_path_obj.read_text().splitlines()[0]
                if first_line.startswith('# '):
                    page_title = first_line.strip('#').strip()
                else:
                    page_title = content_path_obj.stem
            else:
                # File doesn't exist, use filename as title
                page_title = content_path_obj.stem

        include_tag = Tag(name='include', attrs={'path': content_path})

        page_tag = Tag(name='page', attrs={'title': page_title})
        page_tag.append(include_tag)

        if (page_id := tag.get('id', None)) is not None:
            page_tag['id'] = page_id

        include_processor = make_include_preprocessor(parent_folder, process_file)

        try:
            include_processor(include_tag)  # Replaces include_tag with new content
        except FileNotFoundError as e:
            raise FileNotFoundError(f"File not found @ {format_tag(tag)}\n  in {get_file_path(tag)}") from e

        tag.replace_with(page_tag)

    return process_markdown_page



from canvasapi.course import Course

from ..resources import PageInfo


def deploy_page(course: Course, page_info: dict) -> tuple[PageInfo, None]:
    page_id = page_info["canvas_id"]

    if page_id:
        canvas_page = course.get_page(page_id)
        canvas_page.edit(wiki_page=page_info)
    else:
        canvas_page = course.create_page(wiki_page=page_info)

    page_object_info: PageInfo = {
        'id': canvas_page.page_id,
        'title': canvas_page.title,
        'page_url': canvas_page.url,
        'uri': f'/courses/{course.id}/pages/{canvas_page.url}',

        # Following fields have been observed to be missing in some cases
        'url': canvas_page.html_url if hasattr(canvas_page, 'html_url') else None
    }

    return page_object_info, None


def deploy_shell_page(course: Course, page_info: dict) -> tuple[PageInfo, None]:
    shell_page_info = page_info.copy()
    shell_page_info[
        'body'] = "<p>This is a shell page created to break a dependency cycle. The full content will be deployed later.</p>"

    return deploy_page(course, shell_page_info)



from bs4 import Tag

from .attributes import parse_settings, Attribute, parse_date, parse_int
from ..resources import ResourceManager, CanvasResource, get_key
from ..processing_context import get_current_file


def parse_override_tag(override_tag: Tag, parent_type: str, parent_rid: str, resources: ResourceManager):
    """
    Parse an <override> tag that is a child of an assignment or quiz tag.

    Args:
        override_tag: The <override> BeautifulSoup Tag to parse
        parent_type: The type of the parent tag ('assignment' or 'quiz')
        parent_rid: The resource ID (name/title) of the parent assignment or quiz
        resources: The ResourceManager to add the override resource to
    """
    fields = [
        Attribute('available_from', parser=parse_date, new_name='unlock_at'),
        Attribute('available_to', parser=parse_date, new_name='lock_at'),
        Attribute('due_at', parser=parse_date),
        Attribute('late_due', parser=parse_date),
        Attribute('section_id', new_name='course_section_id', required=True, parser=parse_int),
    ]

    settings = {
        "type": "override",
        "assignment_rid": parent_rid,
        "assignment_id": get_key(parent_type, parent_rid, 'id'),
        "rtype": parent_type,
    }

    settings.update(parse_settings(override_tag, fields))

    # Create unique name for this override: parent_rid|section_id
    override_rid = f"{parent_rid}|{settings['course_section_id']}"

    override_resource = CanvasResource(
        type='override',
        id=override_rid,
        data=settings,
        content_path=str(get_current_file().resolve())
    )
    resources.add_resource(override_resource)


def parse_overrides_container(overrides_tag: Tag, parent_type: str, parent_rid: str, resources: ResourceManager):
    """
    Parse an <overrides> container tag that contains multiple <override> child tags.

    Args:
        overrides_tag: The <overrides> BeautifulSoup Tag containing override children
        parent_type: The type of the parent tag ('assignment' or 'quiz')
        parent_rid: The resource ID (name/title) of the parent assignment or quiz
        resources: The ResourceManager to add override resources to
    """
    for override_tag in overrides_tag.findAll('override', recursive=False):
        parse_override_tag(override_tag, parent_type, parent_rid, resources)



from canvasapi.assignment import Assignment
from canvasapi.course import Course

from ..resources import OverrideInfo


def deploy_assignment_override(course: Course, assignment_id: int, override_info: dict) -> tuple[OverrideInfo, None]:
    assignment: Assignment = course.get_assignment(assignment_id)

    if cid := override_info.get('canvas_id'):
        override = assignment.get_override(cid)
        override.edit(assignment_override=override_info)
    else:
        override = assignment.create_override(assignment_override=override_info)

    override_object_info: OverrideInfo = {
        'id': override.id,
    }

    return override_object_info, None


# def deploy_quiz_override(course: Course, quiz_id: int, override_info: dict)-> tuple[AssignmentOverride, str | None]:
#     override_set = course.get_quiz_overrides(**{"quiz_assignment_overrides[][quiz_ids][]": quiz_id})[0]
#
#     if override_set:
#         for override in override_set.overrides:
#             override.edit(**override_info)
#     else:
#         quiz = course.get_quiz(quiz_id)
#         override_set = quiz.create_override(**override_info)
#
#     return override_set, None


def deploy_override(course: Course, override_info: dict) -> tuple[OverrideInfo, None]:
    canvas_id = int(override_info["assignment_id"])
    rtype = override_info["rtype"]

    if rtype == 'quiz':
        # deploy_quiz_override(course, canvas_id, override_info)
        raise TypeError("Override type 'quiz' not currently supported for assignment overrides")

    elif rtype == 'assignment':
        return deploy_assignment_override(course, canvas_id, override_info)

    else:
        # TODO: Try to support Quiz overrides
        raise TypeError(f"{rtype} does not support override")
